/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#define FE_HBR_DISCOVER_DEBUG	FALSE
#define FE_HBR_EVENT_DEBUG		FALSE

#include "houdini/houdini.pmh"

#include <DM/DM_Defines.h>
#include <DM/DM_ViewportType.h>
#include <GR/GR_Detail.h>
#include <GR/GR_DisplayOption.h>
#include <GU/GU_PrimCircle.h>
#include <MSS/MSS_SingleOpState.h>
#include <RE/RE_Render.h>
#include <RE/RE_Material.h>

#define MSS_CLICK_BUTTONS (DM_PRIMARY_BUTTON|DM_SECONDARY_BUTTON)

extern "C"
{

//* discovery point for Houdini
FE_DL_EXPORT void newModelState(BM_ResourceManager *a_pManager)
{
#if FE_HBR_DISCOVER_DEBUG
	feLog("newModelState\n");
#endif

	fe::HoudiniBrush::discoverBrushes(a_pManager);
}

}

namespace fe
{

int HoudiniBrush::discoverBrushes(BM_ResourceManager *a_pManager)
{
	String brushNames=HoudiniSOP::brushNames();

#if FE_HBR_DISCOVER_DEBUG
	feLog("HoudiniBrush::discoverBrushes \"%s\"\n",brushNames.c_str());
#endif

	U32 brushCount=0;
	while(!brushNames.empty())
	{
		const String brushName=brushNames.parse();

#if FE_HBR_DISCOVER_DEBUG
		feLog("HoudiniBrush::discoverBrushes %d \"%s\"\n",
				brushCount,brushName.c_str());
#endif

		a_pManager->registerState(new PI_StateTemplate(
				brushName.c_str(),					// state name
				("FE "+brushName).c_str(),			// English name
				"FE",								// icon name
				(void *)fe::HoudiniBrush::construct,
				NULL,
				PI_VIEWER_SCENE,
				PI_NETMASK_SOP,						// marks as a SOP state
				0));

		brushCount++;
	}

	return brushCount;
}

BM_State* HoudiniBrush::construct(BM_View &a_rView,
	PI_StateTemplate &a_rTemplate,BM_SceneManager *a_pScene)
{
	return new HoudiniBrush((JEDI_View &)a_rView,a_rTemplate,a_pScene);
}

/*
#define DM_VIEWPORT_ORTHO       0x0ff
#define DM_VIEWPORT_TOP         0x001
#define DM_VIEWPORT_FRONT       0x002
#define DM_VIEWPORT_RIGHT       0x004
#define DM_VIEWPORT_BOTTOM      0x008
#define DM_VIEWPORT_BACK        0x010
#define DM_VIEWPORT_LEFT        0x020
#define DM_VIEWPORT_PERSPECTIVE 0xf00
#define DM_VIEWPORT_ALL_3D      (DM_VIEWPORT_PERSPECTIVE|DM_VIEWPORT_ORTHO)
#define DM_VIEWPORT_UV          0x1000
#define DM_VIEWPORT_ALL         (DM_VIEWPORT_ALL_3D | DM_VIEWPORT_UV)
*/

HoudiniBrush::HoudiniBrush(JEDI_View &a_rView,PI_StateTemplate &a_rTemplate,
	BM_SceneManager *a_pScene,char *a_pCursor):
	MSS_SingleOpState(a_rView,a_rTemplate,a_pScene,a_pCursor),
	m_doubleClick(FALSE),
	m_idleQueued(FALSE),
	m_isBrushVisible(false)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::HoudiniBrush %p\n",this);
#endif

	//* only use this state in 3D viewports
	setViewportMask(DM_VIEWPORT_PERSPECTIVE);

	set(m_viewport);

	HoudiniContext houdiniContext;
	sp<Scope> spScope=houdiniContext.scope();
	m_event.bind(spScope);
	sp<Layout> spLayout=m_event.layout();
	Record record=spScope->createRecord(spLayout);
	m_event.bind(record);
	m_event.reset();
}

HoudiniBrush::~HoudiniBrush(void)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::~HoudiniBrush %p\n",this);
#endif
}

const char* HoudiniBrush::className(void) const
{
	return "HoudiniBrush";
}

void HoudiniBrush::brush(void)
{
	HoudiniSOP* pHoudiniSOP=dynamic_cast<HoudiniSOP*>(getNode());

#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::brush %p SOP %p\n",this,pHoudiniSOP);
#endif

	if(!pHoudiniSOP)
	{
		return;
	}

	//* TODO only need to do once
	m_spDrawPerspective=pHoudiniSOP->drawBrush();
	m_spDrawOrtho=pHoudiniSOP->drawBrushOverlay();
	sp<DrawCached> spDrawCached=m_spDrawPerspective;
	if(spDrawCached.isValid())
	{
		spDrawCached->setDrawChain(m_spDrawOpenGL);
	}
	spDrawCached=m_spDrawOrtho;
	if(spDrawCached.isValid())
	{
		spDrawCached->setDrawChain(m_spDrawOpenGL);
	}
//	sp<ViewI> spViewI=m_spDrawOpenGL->view();
//	m_spDrawPerspective->setView(spViewI);
//	m_spDrawOrtho->setView(spViewI);

	const I32 centerX=width(m_viewport)/2;
	const I32 centerY=height(m_viewport)/2;
	if(centerX<1 || centerY<1)
	{
		feLog("HoudiniBrush::brush no viewport (%d %d)\n",
				width(m_viewport),height(m_viewport));
		return;
	}

	UT_Matrix4 brushMatrix;
	getViewportTransform(brushMatrix);
	const float* floats=brushMatrix.data();

	SpatialTransform cameraTransform;
	set(cameraTransform.column(0),floats[0],floats[1],floats[2]);
	set(cameraTransform.column(1),floats[4],floats[5],floats[6]);
	set(cameraTransform.column(2),floats[8],floats[9],floats[10]);
	set(cameraTransform.column(3),floats[12],floats[13],floats[14]);

	I32 mx;
	I32 my;
	m_event.getMousePosition(&mx,&my);

	UT_Vector3 rayorig;
	UT_Vector3 dir;
	mapToWorld(mx,my,dir,rayorig);
	SpatialVector rayOrigin(rayorig[0],rayorig[1],rayorig[2]);
	SpatialVector rayDirection(dir[0],dir[1],dir[2]);

	mapToWorld(centerX,centerY,dir,rayorig);
	SpatialVector centerOrigin(rayorig[0],rayorig[1],rayorig[2]);
	SpatialVector centerDirection(dir[0],dir[1],dir[2]);

/*
	mapToWorld(0,centerY,dir,rayorig);
	SpatialVector leftOrigin(rayorig[0],rayorig[1],rayorig[2]);
	SpatialVector leftDirection(dir[0],dir[1],dir[2]);
	const Real fovx=2.0*acos(dot(centerDirection,leftDirection))/degToRad;
*/

	mapToWorld(centerX,0,dir,rayorig);
	SpatialVector bottomOrigin(rayorig[0],rayorig[1],rayorig[2]);
	SpatialVector bottomDirection(dir[0],dir[1],dir[2]);
	const Real fovy=2.0*acos(dot(centerDirection,bottomDirection))/degToRad;

	FEASSERT(fovy>1.0);
	FEASSERT(fovy<180.0);

	const Real nearPlane=0.1;	//* TODO
	const Real farPlane=1e4;	//* TODO

	sp<ViewI> spView=m_spDrawOrtho->view();
	sp<ViewMode> spViewMode=spView->viewMode();
	spViewMode->setViewport(m_viewport);
	spViewMode->setCamera(cameraTransform);
	spViewMode->setPerspective(fovy,nearPlane,farPlane);

	pHoudiniSOP->brush(&m_gdp,m_event,rayOrigin,rayDirection);
}

int HoudiniBrush::enter(BM_SimpleState::BM_EntryType a_how)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::enter\n");
#endif

	int result=MSS_SingleOpState::enter(a_how);

	//* ask for handleMouseEvent to be called when the mouse moves or a
	//* mouse button is clicked
	wantsLocates(1);
	addClickInterest(MSS_CLICK_BUTTONS);
	updatePrompt();

	//* turn off the highlight so we can see the color we are painting
//	OP_Node *op=getNode();
//	if(op)
//	{
//		op->setHighlight(0);
//	}

	return result;
}

void HoudiniBrush::exit(void)
{
	//* can happen during shutdown
#if FE_HBR_EVENT_DEBUG
	printf("HoudiniBrush::exit\n");
#endif

	//* cleanup
	wantsLocates(0);
	removeClickInterest(MSS_CLICK_BUTTONS);
	m_isBrushVisible=false;
	redrawScene();
	MSS_SingleOpState::exit();
}

void HoudiniBrush::interrupt(BM_SimpleState *a_pState)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::interrupt\n");
#endif

	wantsLocates(0);
	removeClickInterest(MSS_CLICK_BUTTONS);
	m_isBrushVisible=false;
	redrawScene();
	MSS_SingleOpState::interrupt(a_pState);

	return;
}

void HoudiniBrush::resume(BM_SimpleState *a_pState)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::resume\n");
#endif

	MSS_SingleOpState::resume(a_pState);
	wantsLocates(1);
	addClickInterest(MSS_CLICK_BUTTONS);
	updatePrompt();
	m_isBrushVisible=true;
	redrawScene();
//	brush();

	//* turn off the highlight so we can see the color we are painting
//	OP_Node *op=getNode();
//	if(op)
//	{
//		op->setHighlight(0);
//	}

	return;
}

void HoudiniBrush::printEvent(UI_Event* a_pEvent)
{
	feLog("HoudiniBrush::printEvent"
			" state data %d flags 0x%x id %d type 0x%x"
			" value %d %d %d %d\n",
			a_pEvent->state.data,a_pEvent->state.altFlags,
			a_pEvent->state.id,a_pEvent->state.myType,
			a_pEvent->state.values[0],a_pEvent->state.values[1],
			a_pEvent->state.values[2],a_pEvent->state.values[3]);

	feLog(" state type");
	const int stateType=a_pEvent->state.myType;
	switch(stateType)
	{
		case UI_EVENT_NO_EVENT:
			feLog(" no event\n");
			break;
		case UI_EVENT_KEYBOARD:
			feLog(" key\n");
			break;
		case UI_EVENT_FUNCTION_KEY:
			feLog(" function-key\n");
			break;
		case UI_EVENT_ARROW_KEY:
			feLog(" arrow-key\n");
			break;
		case UI_EVENT_MOUSEWHEEL:
			feLog(" wheel\n");
			break;
		case UI_EVENT_DBLCLK:
			feLog(" double-click\n");
			break;
		case UI_EVENT_BUTTON:
			feLog(" button\n");
			break;
		case UI_EVENT_MOTION:
			feLog(" motion\n");
			break;
		default:
			feLog(" other %d\n",stateType);
			break;
	}
}

int HoudiniBrush::handleMouseEvent(UI_Event* a_pEvent)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::handleMouseEvent\n");
#endif

	return handleAnyEvent(a_pEvent);
}

int HoudiniBrush::handleMouseWheelEvent(UI_Event* a_pEvent)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::handleMouseWheelEvent\n");
#endif

	return handleAnyEvent(a_pEvent);
}

bool HoudiniBrush::handleDoubleClickEvent(UI_Event* a_pEvent)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::handleDoubleClickEvent\n");
#endif

	return handleAnyEvent(a_pEvent);
}

int HoudiniBrush::handleArrowEvent(UI_Event* a_pEvent)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::handleArrowEvent\n");
#endif

	return handleAnyEvent(a_pEvent);
}

void HoudiniBrush::handleGeoChangedEvent(UI_Event* a_pEvent)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::handleGeoChangedEvent\n");
#endif

	handleAnyEvent(a_pEvent);
}

int HoudiniBrush::handleKeyTypeEvent(UI_Event* a_pEvent,BM_Viewport&)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::handleKeyTypeEvent\n");
#endif

	return handleAnyEvent(a_pEvent);
}

int HoudiniBrush::handleKeyEvent(UI_Event* a_pEvent,BM_Viewport&)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::handleKeyEvent\n");
#endif

	return handleAnyEvent(a_pEvent);
}

int HoudiniBrush::handleTransitoryKey(UI_Event* a_pEvent)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::handleTransitoryKey\n");
#endif

	return handleAnyEvent(a_pEvent);
}

void HoudiniBrush::handleEvent(UI_Event* a_pEvent)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::handleEvent\n");
#endif

	handleAnyEvent(a_pEvent);
}

int HoudiniBrush::handleAnyEvent(UI_Event* a_pEvent)
{
	SOP_Node *sop=(SOP_Node *)getNode();
	if(!sop)
	{
#if FE_HBR_EVENT_DEBUG
		feLog("HoudiniBrush::handleAnyEvent not a SOP_Node\n");
#endif
		//* consumed but useless
		return 1;
	}

	const float t=getTime();
	const int x=a_pEvent->state.values[X];
	const int y=a_pEvent->state.values[Y];

#if FE_HBR_EVENT_DEBUG
	printEvent(a_pEvent);
#endif

	WindowEvent::Item modifiers=WindowEvent::e_itemNull;
	if(a_pEvent->state.altFlags&UI_SHIFT_KEY)
	{
#if FE_HBR_EVENT_DEBUG
		feLog(" state SHIFT\n");
#endif
		modifiers=WindowEvent::Item(modifiers|WindowEvent::e_keyShift);
	}
	if(a_pEvent->state.altFlags&UI_CTRL_KEY)
	{
#if FE_HBR_EVENT_DEBUG
		feLog(" state CTRL\n");
#endif
		modifiers=WindowEvent::Item(modifiers|WindowEvent::e_keyControl);
	}
#if FE_HBR_EVENT_DEBUG
	if(a_pEvent->state.altFlags&UI_ALT_KEY)
	{
		feLog(" state ALT\n");
	}
	if(a_pEvent->state.altFlags&UI_COMMAND_KEY)
	{
		feLog(" state COMMAND\n");
	}
#endif

	BWORD send=FALSE;
	BWORD validPosition=TRUE;

	if(a_pEvent->state.myType==UI_EVENT_NO_EVENT)
	{
		//* not supposed to have key modifiers on system events
//		modifiers=WindowEvent::Item(m_event.item()&WindowEvent::e_keyModifiers);

		m_event.setSIS(WindowEvent::e_sourceSystem,
				WindowEvent::Item(WindowEvent::e_itemIdle|modifiers),
				WindowEvent::e_stateNull);
		send=TRUE;
		m_idleQueued=FALSE;
		validPosition=FALSE;
	}

	if(a_pEvent->state.myType==UI_EVENT_KEYBOARD)
	{
		m_event.setSIS(WindowEvent::e_sourceKeyboard,
				WindowEvent::Item(a_pEvent->state.data|modifiers),
				WindowEvent::e_statePress);
		send=TRUE;
	}

	if(a_pEvent->state.myType==UI_EVENT_FUNCTION_KEY)
	{
		m_event.setSIS(WindowEvent::e_sourceKeyboard,
				WindowEvent::Item(
				(WindowEvent::e_keyFunction+a_pEvent->state.data-68)|modifiers),
				WindowEvent::e_statePress);
		send=TRUE;
	}

	if(a_pEvent->state.myType==UI_EVENT_ARROW_KEY)
	{
		m_event.setSIS(WindowEvent::e_sourceKeyboard,
				WindowEvent::Item(
				(WindowEvent::e_keyCursor+a_pEvent->state.data)|modifiers),
				WindowEvent::e_statePress);
		send=TRUE;
	}

	if(a_pEvent->state.myType==UI_EVENT_MOUSEWHEEL)
	{
		m_event.setSIS(WindowEvent::e_sourceMouseButton,
				WindowEvent::Item(WindowEvent::e_itemWheel|modifiers),
				(WindowEvent::State)a_pEvent->state.values[3]);
		send=TRUE;
	}

	if(a_pEvent->state.myType==UI_EVENT_BUTTON)
	{
		const WindowEvent::State state2=m_doubleClick?
				WindowEvent::e_state2ClickDouble:
				WindowEvent::e_state2ClickSingle;
		send=TRUE;

		switch(a_pEvent->state.values[W])
		{
			case 0:
				if(m_event.mouseButtons()==WindowEvent::e_mbLeft)
				{
					m_event.setSIS(WindowEvent::e_sourceMouseButton,
							WindowEvent::Item(
							WindowEvent::e_itemLeft|modifiers),
							WindowEvent::e_stateRelease);
				}
				else if(m_event.mouseButtons()==WindowEvent::e_mbMiddle)
				{
					m_event.setSIS(WindowEvent::e_sourceMouseButton,
							WindowEvent::Item(
							WindowEvent::e_itemMiddle|modifiers),
							WindowEvent::e_stateRelease);
				}
				else
				{
					send=FALSE;
				}
				m_event.setMouseButtons(WindowEvent::e_mbNone);
				m_doubleClick=FALSE;
				break;
			case DM_PRIMARY_BUTTON:
				m_event.setSIS(WindowEvent::e_sourceMouseButton,
						WindowEvent::Item(WindowEvent::e_itemLeft|modifiers),
						WindowEvent::e_statePress);
				m_event.setMouseButtons(WindowEvent::e_mbLeft);
				break;
			case DM_SECONDARY_BUTTON:
				m_event.setSIS(WindowEvent::e_sourceMouseButton,
						WindowEvent::Item(WindowEvent::e_itemMiddle|modifiers),
						WindowEvent::e_statePress);
				m_event.setMouseButtons(WindowEvent::e_mbMiddle);
				break;
		}
		m_event.setState2(state2);
	}
	else
	{
		m_doubleClick=FALSE;
		m_event.setState2(WindowEvent::e_stateNull);
	}

	if(a_pEvent->state.myType==UI_EVENT_DBLCLK)
	{
		m_doubleClick=TRUE;
	}

	if(!m_idleQueued)
	{
		HoudiniSOP* pHoudiniSOP=dynamic_cast<HoudiniSOP*>(getNode());
		if(pHoudiniSOP)
		{
			String idle=pHoudiniSOP->idleEvents();
			if(idle=="any" || (m_isBrushVisible && idle=="viewer"))
			{
				const UI_EventType eventType=UI_EVENT_NO_EVENT;
				UI_Object* target=this;
				const int purge=0;	//* what does this affect?

				generateEvent(eventType,target,purge);
				m_idleQueued=TRUE;
			}
		}
	}

	if(validPosition)
	{
		m_event.setMousePosition(x,y);
	}

	if(send)
	{
		brush();
		return 1;
	}

#if FE_HBR_EVENT_DEBUG
	switch(a_pEvent->type)
	{
		case UI_EVENT_VALUE_CHANGE:
			feLog(" type value_change\n");
			break;
		default:
			feLog(" type other %d\n",a_pEvent->type);
			break;
	}
	switch(a_pEvent->reason)
	{
		case UI_VALUE_START:
			feLog(" reason start\n");
			break;
		case UI_VALUE_CHANGED:
			feLog(" reason changed\n");
			break;
		case UI_VALUE_LOCATED:
			feLog(" reason located\n");
			break;
		case UI_VALUE_PICKED:
			feLog(" reason picked\n");
			break;
		case UI_VALUE_ACTIVE:
			feLog(" reason active\n");
			break;
		default:
			feLog(" reason other %d\n",a_pEvent->reason);
			break;
	}
#endif

#if FALSE
	I32 oldX;
	I32 oldY;
	m_event.getMousePosition(&oldX,&oldY);
	if(x==oldX && y==oldY)
	{
		return 1;
	}
#endif

	if(a_pEvent->reason==UI_VALUE_LOCATED)
	{
		m_event.setSIS(WindowEvent::e_sourceMousePosition,
				WindowEvent::Item(WindowEvent::e_itemMove|modifiers),
				WindowEvent::e_stateRepeat);
	}
	else
	{
		//* trigger a cook of the CustomBrush SOP so it can cook with
		//* the current stroke values
//		sop->getCookedGeo(context);

		m_event.setSIS(WindowEvent::e_sourceMousePosition,
				WindowEvent::Item(WindowEvent::e_itemDrag|modifiers),
				WindowEvent::e_stateRepeat);

	}

	updateBrush(x,y);
	brush();

	return 1;
}

U32 HoudiniBrush::initFont(void)
{
	//* HACK load font manually
	const String name="-*-clean-medium-r-*-*-8-*-*-*-*-*-*-*";
	Display* pDisplayHandle=XOpenDisplay(NULL);
	XFontStruct* pFontStruct=XLoadQueryFont(pDisplayHandle,name.c_str());
	const U32 lists=pFontStruct->max_char_or_byte2+1;
	const U32 fontBase=glGenLists(lists);

	const Font fontid=pFontStruct->fid;
	const U32 first=pFontStruct->min_char_or_byte2;
	const U32 last=pFontStruct->max_char_or_byte2;

	glXUseXFont(fontid,first,last-first+1,fontBase+first);

	return fontBase;
}

//* NOTE called twice per refresh: ghost 1, ghost 0 (then geometry render)
void HoudiniBrush::doRender(RE_Render *a_pRender,short a_x,short a_y,
	int a_ghost)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::doRender %d %d xy %d %d ghost %d\n",
			isPreempted(),m_isBrushVisible,a_x,a_y,a_ghost);
#endif

	if(a_ghost || isPreempted() || !m_isBrushVisible)
	{
		return;
	}

	UT_DimRect rect=a_pRender->getViewport2DI();
	set(m_viewport,rect.x(),rect.y(),rect.w(),rect.h());

//	GR_Detail rgdp;
//	rgdp.wireDraw(&m_gdp,*a_pRender,getViewportLOD(),1);

#if TRUE
//	a_pRender->lockContextForRender();

	//* sandbox any OpenGL state changes
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	if(m_spDrawOpenGL.isNull())
	{
		HoudiniContext houdiniContext;
		m_spDrawOpenGL=houdiniContext.master()->registry()->create(
				"DrawI.DrawOpenGL");

		const U32 fontBase=initFont();
		m_spDrawOpenGL->setFontBase(fontBase);

		sp<DrawMode> spDrawMode(new DrawMode);
		spDrawMode->setDrawStyle(DrawMode::e_solid);
		spDrawMode->setLineWidth(3.0);
		spDrawMode->setPointSize(8.0);
//		spDrawMode->setAntialias(TRUE);

		m_spDrawOpenGL->pushDrawMode(spDrawMode);
	}

	sp<DrawMode> spDrawMode=m_spDrawOpenGL->popDrawMode();
	m_spDrawOpenGL->pushDrawMode(spDrawMode);

	//* perspective

	sp<ViewI> spViewI=m_spDrawOpenGL->view();
//	spViewI->use(ViewMode::e_perspective);

	if(m_spDrawPerspective.isValid())
	{
		m_spDrawOpenGL->pushDrawMode(spDrawMode);
//		feLog("operator 3D\n");
		m_spDrawPerspective->flushLive();
		m_spDrawOpenGL->popDrawMode();
	}

#if FALSE
	feLog("axes 3D\n");
	m_spDrawOpenGL->drawAxes(1.0);

	SpatialTransform xform;
	setIdentity(xform);
	translate(xform,SpatialVector(0.2,0.3,0.4));
	rotate(xform,-90.0*degToRad,e_xAxis);

	const SpatialVector scale(0.1,0.1,0.1);
	const Color magenta(1.0,0.0,1.0,1.0);

	feLog("sphere 3D\n");
	m_spDrawOpenGL->drawSphere(xform,&scale,magenta);

	feLog("circle 3D\n");
	translate(xform,SpatialVector(0.2,0.3,0.0));
	m_spDrawOpenGL->drawCircle(xform,&scale,magenta);
#endif

	//* ortho

	sp<ViewMode> spViewMode=spViewI->viewMode();
	spViewMode->setViewport(m_viewport);
	spViewI->use(ViewMode::e_ortho);

	if(m_spDrawOrtho.isValid())
	{
		m_spDrawOpenGL->pushDrawMode(spDrawMode);
//		feLog("operator 2D\n");
		m_spDrawOrtho->flushLive();
		m_spDrawOpenGL->popDrawMode();
	}

#if FALSE
	const Color yellow(1.0f,1.0f,0.0f,1.0f);
	const SpatialVector circleScale(10.0f,10.0f,10.0f);
	const SpatialVector center(30.0,60.0,1.0f);

	feLog("circle 2D\n");
	SpatialTransform transform;
	setIdentity(transform);
	translate(transform,center);
	m_spDrawOpenGL->drawCircle(transform,&circleScale,yellow);

	const Color cyan(0.0f,1.0f,1.0f,1.0f);
	m_spDrawOpenGL->drawAlignedText(
			SpatialVector(100.0f,120.0f,1.0f),"Hello",cyan);

	const Color darkred(0.5f,0.0f,0.0f,1.0f);
	const SpatialVector dot(30.0,70.0,1.0);
	m_spDrawOpenGL->drawPoints(&dot,NULL,1,FALSE,&darkred);
#endif

	//* restore original state
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glPopAttrib();

//	a_pRender->unlockContextAfterRender();
#endif
}

void HoudiniBrush::updatePrompt(void)
{
#if FE_HBR_EVENT_DEBUG
	feLog("HoudiniBrush::updatePrompt\n");
#endif

	HoudiniSOP* pHoudiniSOP=dynamic_cast<HoudiniSOP*>(getNode());
	const String prompt=pHoudiniSOP? pHoudiniSOP->prompt():
			"insert helpful brush instructions here";

	showPrompt(prompt.c_str());
//	showPrompt("LMB to apply stroke.  MMB to erase."
//			"  Shift-LMB to adjust radius.");
}

void HoudiniBrush::updateBrush(int a_x,int a_y)
{
//	feLog("HoudiniBrush::updateBrush\n");

	//* ensure the brush is visible
	m_isBrushVisible=true;
//	redrawScene();
}

} // namespace
