/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_FCO_DEBUG		TRUE

#include "operator/operator.pmh"

using namespace fe;

void FollicleOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<Real>("pickRadius")=1.0;
	catalog<String>("pickRadius","label")="Pick Radius";

	catalog<I32>("lineWidth")=2;
	catalog<I32>("lineWidth","min")=1;
	catalog<I32>("lineWidth","max")=10;

	catalog<I32>("rimSize")=4;
	catalog<I32>("rimSize","min")=2;
	catalog<I32>("rimSize","max")=10;

	catalog< sp<Component> >("Brush");
	catalog<String>("Brush","implementation")="DrawI";
	catalog<String>("Brush","prompt")="prompt here";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Driver Surface");

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
//	catalog<bool>("Output Surface","recycle")=true;

	m_spWireframe=new DrawMode();
	m_spWireframe->setDrawStyle(DrawMode::e_wireframe);
	m_spWireframe->setLineWidth(2.0);

	m_spSolid=new DrawMode();
	m_spSolid->setDrawStyle(DrawMode::e_solid);
	m_spSolid->setBackfaceCulling(FALSE);
}

void FollicleOp::handle(Record& a_rSignal)
{
	const Real pickRadius=catalog<Real>("pickRadius");
	const I32 lineWidth=catalog<I32>("lineWidth");
	const I32 rimSize=catalog<I32>("rimSize");

	sp<DrawI> spDrawBrush;
	accessBrush(spDrawBrush,a_rSignal,e_quiet);

	sp<DrawI> spDrawOverlay;
	accessBrushOverlay(spDrawOverlay,a_rSignal,e_quiet);

#if FE_FCO_DEBUG
	feLog("FollicleOp::handle brush %d brushed %d\n",
			spDrawBrush.isValid(),m_brushed);
#endif

	if(spDrawBrush.isValid())
	{
		m_spWireframe->setLineWidth(lineWidth);

		spDrawBrush->pushDrawMode(m_spWireframe);
		spDrawOverlay->pushDrawMode(m_spWireframe);

		m_brushed=TRUE;

		const Color white(1.0,1.0,0.8,1.0);
		const Color grey(0.8,0.8,1.0,1.0);
		const Color red(0.8,0.0,0.0,1.0);
		const Color green(0.0,0.5,0.0,1.0);
		const Color blue(0.0,0.0,1.0,1.0);
		const Color cyan(0.0,0.8,1.0,1.0);
		const Color magenta(0.8,0.0,0.8,1.0);
		const Color yellow(1.0,1.0,0.0,1.0);
		const Color blank(0.0,0.0,0.0,0.0);

		const SpatialVector up(0.0,1.0,0.0);

		const Real sphereRadius=0.02;
		const SpatialVector sphereScale(sphereRadius,sphereRadius,sphereRadius);

		if(m_spDriver.isValid())
		{
			const SpatialVector& rRayOrigin=rayOrigin(a_rSignal);
			const SpatialVector& rRayDirection=rayDirection(a_rSignal);

			m_event.bind(windowEvent(a_rSignal));

#if FE_FCO_DEBUG
			feLog("%s\n",c_print(m_event));
#endif

			SpatialVector intersection;
			set(intersection);

			const Real maxDistance=100.0;
			sp<SurfaceI::ImpactI> spImpact=m_spDriver->rayImpact(
					rRayOrigin,rRayDirection,maxDistance);
			if(spImpact.isValid())
			{
				sp<SurfaceTriangles::Impact> spTriImpact=spImpact;
				if(spTriImpact.isValid())
				{
					const I32 face=spTriImpact->face();
					const I32 pickIndex=spTriImpact->pickIndex();
					intersection=spTriImpact->intersection();
					const SpatialVector hitNormal=
							spTriImpact->normal();
#if FE_FCO_DEBUG
					feLog("FollicleOp::handle face %d pick %d %s\n",
							face,pickIndex,c_print(intersection));
#endif

					const I32 depth=0;
					drawPicker(spDrawBrush,intersection,up,hitNormal,
							pickRadius,depth,&yellow);
				}
			}

			sp<ViewI> spView=spDrawOverlay->view();
			sp<ViewMode> spViewMode=spView->viewMode();
			SpatialTransform camera=spViewMode->camera();
			SpatialTransform invCamera;
			invert(invCamera,camera);

//			feLog("cam\n%s\n",c_print(camera));
//			feLog("inv\n%s\n",c_print(invCamera));

			const SpatialVector deltaX=invCamera.column(0);
			const SpatialVector deltaY=invCamera.column(1);

			I32 closest= -1;
			Real closestDist= -1.0;

			const U32 outputCount=m_spOutputPoint->count();
			m_colorArray.resize(outputCount);

			const Real rad2=pickRadius*pickRadius;
			for(I32 index=0;index<outputCount;index++)
			{
				const SpatialVector normal=
						m_spOutputNormal->spatialVector(index);
				const Real hitDot=dot(normal,invCamera.column(2));
				if(hitDot<0)
				{
					m_colorArray[index]=blank;
					continue;
				}

				const SpatialVector point=
						m_spOutputPoint->spatialVector(index);

				m_colorArray[index]=green;

				if(spImpact.isValid())
				{
					const Real dist2=magnitudeSquared(point-intersection);
					if(dist2<rad2)
					{
						m_colorArray[index]=blue;

						const Real dist=sqrt(dist2);
						if(closest<0 || closestDist>dist)
						{
							closest=index;
							closestDist=dist;
						}
					}
				}
			}

#if FALSE
			spDrawBrush->pushDrawMode(m_spSolid);
			spDrawBrush->draw(m_spDriver,&white);
			spDrawBrush->popDrawMode();
#endif

			if(m_event.isKeyPress(WindowEvent::e_itemAny))
			{
				U32 item=m_event.item();
				BWORD shifted=FALSE;
				if(item&WindowEvent::e_keyShift)
				{
					shifted=TRUE;
					item^=WindowEvent::e_keyShift;
				}

				if(item==WindowEvent::e_keyDelete)
				{
				}
			}

			const WindowEvent::MouseButtons buttons=m_event.mouseButtons();
			if(buttons)
			{
				U32 item=m_event.item();
				BWORD shifted=FALSE;
				if(item&WindowEvent::e_keyShift)
				{
					shifted=TRUE;
					item^=WindowEvent::e_keyShift;
				}
				BWORD ctrled=FALSE;
				if(item&WindowEvent::e_keyControl)
				{
					ctrled=TRUE;
					item^=WindowEvent::e_keyControl;
				}

				const I32 mouseX=m_event.mouseX();
				const I32 mouseY=m_event.mouseY();

				const I32 moveX=mouseX-m_lastX;
				const I32 moveY=mouseY-m_lastY;

				const Real moveScale=3e-3;	//* TODO screen aware

				const SpatialVector delta=moveScale*(deltaX*moveX+deltaY*moveY);

				if(buttons&WindowEvent::e_mbLeft)
				{
					if(m_event.isMouseDrag())
					{
						if(closest>=0)
						{
							movePicked(delta);
						}
					}
					else
					{
						if(closest>=0)
						{
							if(ctrled)
							{
								pickToggle(closest);
							}
							else if(!picked(closest))
							{
								m_pickArray.clear();
								m_pickArray.push_back(closest);
							}
						}
						else if(!ctrled)
						{
							m_pickArray.clear();
						}
					}
				}

				if(buttons&WindowEvent::e_mbMiddle)
				{
					if(m_event.isMouseDrag())
					{
					}
				}

				m_lastX=mouseX;
				m_lastY=mouseY;
			}

		}

		const I32 outputCount=m_spOutputPoint->count();
		const I32 pickCount=m_pickArray.size();
		for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
		{
			const I32 index=m_pickArray[pickIndex];
			if(index>=0 && index<outputCount)
			{
				m_colorArray[index]=yellow;
			}
		}

		for(I32 index=0;index<outputCount;index++)
		{
			const Color& rColor=m_colorArray[index];
			if(rColor[3]<=0.0)
			{
				continue;
			}

			const SpatialVector point=
					m_spOutputPoint->spatialVector(index);

			drawFollicle(spDrawOverlay,point,rimSize,rColor);
		}

		spDrawOverlay->popDrawMode();
		spDrawBrush->popDrawMode();

		return;
	}

	if(!access(m_spInputPoint,"Input Surface",e_point,e_position)) return;

	if(!accessOutput(m_spOutputPoint,a_rSignal,e_point,e_position)) return;

	if(!accessOutput(m_spOutputNormal,a_rSignal,e_point,e_normal)) return;

	const BWORD paramChanged=!m_brushed;
	m_brushed=FALSE;

	if(paramChanged)
	{
		if(!access(m_spDriver,"Driver Surface")) return;
	}

	const BWORD rebuild=(paramChanged ||
			catalog<bool>("Output Surface","replaced"));

	sp<RecordArray> spRA=m_spState->getArray("Move");
	if(spRA.isValid())
	{
		for(int i=0;i<spRA->length();i++)
		{
			const I32 primIndex=m_aIndex(spRA,i);

			Record edit=spRA->getRecord(i);

			const U32 subCount=m_spInputPoint->subCount(primIndex);
			if(subCount)
			{
				for(U32 subIndex=0;subIndex<subCount;subIndex++)
				{
					const SpatialVector inputPoint=
							m_spInputPoint->spatialVector(primIndex,subIndex);

					const SpatialVector outputPoint=
							inputPoint+m_aLocation(edit);

					m_spOutputPoint->set(primIndex,subIndex,outputPoint);
				}
			}
		}
	}

	return;
}

BWORD FollicleOp::pickToggle(I32 a_index)
{
	const I32 pickCount=m_pickArray.size();
	for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
	{
		if(a_index==m_pickArray[pickIndex])
		{
			m_pickArray.erase(m_pickArray.begin()+pickIndex);
			return TRUE;
		}
	}
	m_pickArray.push_back(a_index);
	return FALSE;
}

BWORD FollicleOp::picked(I32 a_index)
{
	const I32 pickCount=m_pickArray.size();
	for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
	{
		if(a_index==m_pickArray[pickIndex])
		{
			return TRUE;
		}
	}
	return FALSE;
}

Record FollicleOp::getEdit(I32 a_index)
{
	Record edit=m_spEditMap->lookup(a_index);
	if(!edit.isValid())
	{
		edit=m_spScope->createRecord("Move");
		m_aIndex(edit)=a_index;

		m_spState->add(edit);
	}

	return edit;
}

void FollicleOp::movePicked(const SpatialVector& a_rDelta)
{
	const I32 pickCount=m_pickArray.size();
	for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
	{
		const I32 index=m_pickArray[pickIndex];
		const SpatialVector inputPoint=m_spInputPoint->spatialVector(index);
		Record edit=getEdit(index);
		m_aLocation(edit)+=a_rDelta;
		const SpatialVector outputPoint=inputPoint+m_aLocation(edit);

		snapToSurface(index);
	}
}

void FollicleOp::snapToSurface(I32 a_index)
{
	const SpatialVector point=m_spInputPoint->spatialVector(a_index);
	Record edit=getEdit(a_index);

	sp<SurfaceI::ImpactI> spImpact=
			m_spDriver->nearestPoint(point+m_aLocation(edit));
	if(!spImpact.isValid())
	{
		return;
	}
	const SpatialVector intersection=spImpact->intersection();

	m_aLocation(edit)=intersection-point;
}

void FollicleOp::drawFollicle(sp<DrawI>& a_rspDraw,
	const SpatialVector& a_rPoint,Real a_rimSize,const Color& a_rColor)
{
	sp<ViewI> spView=a_rspDraw->view();

	const Vector2i projected=spView->project(a_rPoint,ViewMode::e_perspective);

	const Vector2i rim(a_rimSize,a_rimSize);
	Box2i box=projected-rim;
	box.size()=2.0*rim+Vector2i(1,1);

	a_rspDraw->drawBox(box,a_rColor);
}

void FollicleOp::setupState(void)
{
	m_aIndex.populate(m_spScope,"Move","Index");
	m_aLocation.populate(m_spScope,"Move","Location");

	m_aIndex.populate(m_spScope,"Add","Index");
	m_aLocation.populate(m_spScope,"Add","Location");
}

BWORD FollicleOp::loadState(const String& rBuffer)
{
	if(OperatorState::loadState(rBuffer))
	{
		m_spEditMap=m_spState->lookupMap(m_aIndex);
		return TRUE;
	}

	return FALSE;
}
