/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_OTHR_DEBUG			FALSE
#define FE_OTHR_THREAD_VERBOSE	FALSE

#include "operator/operator.pmh"

using namespace fe;

OperatorThreaded::~OperatorThreaded(void)
{
	finish();
}

void OperatorThreaded::finish(void)
{
	Range stop;
	for(U32 m=0;m<m_threadsAlive;m++)
	{
		m_spGang->post(stop);
	}
	if(m_spGang->workers())
	{
		m_spGang->finish();
	}
	m_threadsAlive=0;
}

void OperatorThreaded::initialize(void)
{
	catalog<I32>("Threads")=2;
	catalog<I32>("Threads","min")=1;
	catalog<I32>("Threads","high")=16;
	catalog<bool>("Threads","joined")=true;
	catalog<String>("Threads","IO")="input output";
	catalog<String>("Threads","hint")=
			"Number of workers to process the data in parallel."
			"  The number of jobs may not be the same."
			"  When the thread count is set to one,"
			" no threads are spawned and the entire"
			" processing job is done at once with the existing thread.";

	catalog<bool>("AutoThread")=true;
	catalog<bool>("AutoThread","joined")=true;
	catalog<String>("AutoThread","IO")="input output";
	catalog<String>("AutoThread","hint")=
			"Automatically set the number of threads.";

	catalog<bool>("StayAlive")=false;
	catalog<String>("StayAlive","label")="Stay Alive";
	catalog<String>("StayAlive","hint")=
			"Keep the threads running and looking for jobs even"
			" when this node isn't being processed."
			"  This may speed up quick nodes where launching threads is"
			" a large portion of the processing time,"
			" but this technique should be considered tempermental"
			" (crashes may become more likely).";

	m_spGang=new Gang<Worker,Range>();
}

void OperatorThreaded::handle(Record& a_rSignal)
{
#if FE_OTHR_DEBUG
	feLog("OperatorThreaded::handle \"%s\"\n",name().c_str());
#endif

	BWORD& rAutoThread=catalog<bool>("AutoThread");
	I32& rThreads=catalog<I32>("Threads");
	const bool stayAlive=(catalog<bool>("StayAlive") && rThreads>1);

	if(m_threadsAlive &&
			(!stayAlive || (stayAlive && m_threadsAlive!=rThreads)))
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Pre-finish stayAlive %d threadsAlive %d\n",
				stayAlive,m_threadsAlive);
#endif
		finish();
	}

	if(rAutoThread)
	{
		rThreads=boost::thread::hardware_concurrency();
	}

	//* for a single thread, run directly
	if(rThreads==1)
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("OperatorThreaded::handle single thread [%d,%d]\n",
				m_fullRange.m_start,m_fullRange.m_end);
#endif
		run(-1,m_fullRange);
		return;
	}

	const U32 count=m_fullRange.m_end-m_fullRange.m_start+1;

	const U32 jobs=rThreads*4;	//* TODO param?
	const U32 each=(count>jobs)? (count+jobs-1)/jobs: 1;

	Range range;
	for(U32 m=0;m<count;m+=each)
	{
		range.m_start=m_fullRange.m_start+m;
		range.m_end=range.m_start+each-1;
		if(range.m_end>m_fullRange.m_end)
		{
			range.m_end=m_fullRange.m_end;
		}

#if FE_OTHR_THREAD_VERBOSE
		feLog("Post [%d,%d]\n",range.m_start,range.m_end);
#endif
		if(range.m_end>=range.m_start)
		{
			m_spGang->post(range);
		}
	}

	if(!stayAlive || !m_threadsAlive)
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Start %d threads\n",rThreads);
#endif
		if(!m_spGang->start(rThreads,sp<OperatorThreaded>(this)))
		{
			feLog("OperatorThreaded::handle \"%s\" threading failed"
					" -> switching to single thread\n",name().c_str());

			m_threadsAlive=0;
			rAutoThread=FALSE;
			rThreads=1;
			run(-1,m_fullRange);
			return;
		}
		if(stayAlive)
		{
			m_threadsAlive=rThreads;
		}
	}

	if(!stayAlive)
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Post-finish stayAlive %d threadsAlive %d\n",
				stayAlive,m_threadsAlive);
#endif
		finish();
	}

#if FE_OTHR_DEBUG
	feLog("OperatorThreaded::handle done\n");
#endif
}

void OperatorThreaded::Worker::operator()(void)
{
#if FE_OTHR_THREAD_VERBOSE
	feLog("Starting thread %d\n",m_id);
#endif
	const BWORD& wait=m_spOperatorThreaded->m_threadsAlive;
	BWORD done=FALSE;
	while(!done)
	{
		Range range;

		if(wait)
		{
#if FE_OTHR_THREAD_VERBOSE
			feLog("Thread %d waiting\n",m_id);
#endif
			m_spJobQueue->waitForJob(range);
		}
		else
		{
			done=!m_spJobQueue->take(range);
		}

#if FE_OTHR_THREAD_VERBOSE
		feLog("Thread %d Job [%d,%d]\n",m_id,range.m_start,range.m_end);
#endif

		if(range.m_start<0)
		{
			done=TRUE;
		}
		else if(!done)
		{
			m_spOperatorThreaded->run(m_id,range);
		}
	}
#if FE_OTHR_THREAD_VERBOSE
	feLog("Sync thread %d\n",m_id);
#endif
	U32 leader=m_spJobQueue->synchronize();
#if FE_OTHR_THREAD_VERBOSE
	feLog("Sync'd thread %d leader=%d\n",m_id,leader);
#endif
}
